@@ -23,44 +23,114 @@ from __future__ import annotations  # noqa: F407
 
 import base64
 import datetime
+import decimal
 import getpass
 import io
 import json
 import logging
 import os
+import pickle as pkl
 import re
 import sys
 import tarfile
 import tempfile
 import textwrap
 import time
+import typing as t
+import warnings
 import webbrowser
 from dataclasses import dataclass
 from io import BytesIO
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import Any, Dict, List, Optional
 
-import cloudpickle
 import matplotlib.pyplot as plt
 import numpy as np
+import pandas as pd
 import pyarrow as pa
 import pyarrow.parquet as pq
+import sarus_data_spec
 from PIL import Image
 from requests import Session
 
-import pandas as pd
-import sarus.pandas as spd
-import sarus.tensorflow
-import tensorflow as tf
-from tensorflow.keras.models import Model
+from sarus.dataspec_wrapper import DataSpecWrapper
+from sarus.manager.typing import LocalSDKManager
+from sarus.typing import DataSpecVariant, SyncPolicy
+
+with warnings.catch_warnings():
+    warnings.simplefilter("ignore")
+    import sarus_data_spec.typing as st
+    from sarus_data_spec.context import push_global_context
+    from sarus_data_spec.protobuf.utilities import (
+        deserialize,
+        dict_deserialize,
+        serialize,
+    )
+    from sarus_data_spec.query_manager.simple_rules import attach_variant
+    from sarus_data_spec.variant_constraint import variant_constraint
+
+    from sarus.context.local_sdk import LocalSDKContext
+    from sarus.pandas.dataframe import DataFrame
+
+    try:
+        import tensorflow as tf
+
+        import sarus.legacy.tensorflow as sltf
+    except ModuleNotFoundError:
+        pass  # error message in sarus_data_spec.typing
 
 
 @dataclass
 class Synthetic:
+    """Simple class to represent synthetic data."""
+
     data: Optional[pa.Table] = None
     rows: Optional[int] = None
 
 
-class Dataset:
+class DatasetList(list):
+    """Custom class to lazily fetch a Dataset when indexing."""
+
+    def __init__(
+        self, iterable: t.Iterable[ShallowDataset], client: Client
+    ) -> None:
+        super().__init__(iterable)
+        self.client = client
+
+    def __getitem__(self, i: int) -> Dataset:
+        shallow_dataset = super().__getitem__(i)
+        slugname = shallow_dataset.name
+        return self.client.dataset(slugname=slugname)
+
+    def __setitem__(self, i: int, value: t.Any) -> None:
+        raise NotImplementedError("A DatasetList is immutable.")
+
+    def append(self, value: t.Any) -> None:
+        raise NotImplementedError("A DatasetList is immutable.")
+
+    def insert(self, i: int, value: t.Any) -> None:
+        raise NotImplementedError("A DatasetList is immutable.")
+
+    def extend(self, other: DatasetList) -> None:
+        raise NotImplementedError("A DatasetList is immutable.")
+
+
+class ShallowDataset:
+    """Shallow dataset that is merely used as a placeholder for slugname."""
+
+    def __init__(self, id: int, slugname: str):
+        self.id = id
+        self.name = slugname
+
+    def __repr__(self) -> str:
+        return f"<Sarus Dataset slugname={self.name} id={self.id}>"
+
+    @classmethod
+    def _from_dict(cls, data: dict) -> ShallowDataset:
+        # `data` is the JSON returned by calling GET /datasets/
+        return cls(data.get("id"), data.get("name"))
+
+
+class Dataset(DataSpecWrapper[t.Iterator[pa.RecordBatch]]):
     """A class representing a Sarus Dataset.
 
     This class is the interface to the protected data. It enables to inspect the
@@ -69,12 +139,10 @@ class Dataset:
     queries.
 
     Args:
-        slugname: The dataset slugname.
         id: The dataset id.
         client: The Sarus client where the dataset is defined.
         type_metadata: A serialized json holding the dataset metadata.
         marginals: A serialized json holding the dataset marginals.
-        URI: The dataset URI.
         human_description: A short human readable description.
         policy:
         transforms:
@@ -83,24 +151,25 @@ class Dataset:
 
     def __init__(
         self,
-        slugname: str,
         id: int,
         client: "Client",
+        dataspec: st.DataSpec,
+        is_bigdata: Optional[bool] = None,
         type_metadata: Optional[str] = None,
-        URI: Optional[str] = None,
         human_description: Optional[str] = None,
         marginals: Optional[str] = None,
         policy: Optional[dict] = None,
         synthetic: Dict[str, Synthetic] = None,
         transforms: List = [],
     ):
-        self.name: str = slugname
         self.id: int = id
         self.client: "Client" = client
+        self.is_bigdata: Optional[bool] = is_bigdata
+        self.name = dataspec.properties().get("slugname")
+        self._set_dataspec(dataspec)
         self.type_metadata: Optional[Dict[str, Any]] = None
         if type_metadata is not None:
             self.type_metadata = json.loads(type_metadata)
-        self.URI: Optional[str] = URI
         self.human_description: Optional[str] = human_description
         if synthetic is None:
             self._synthetic = dict(original=Synthetic(), encoded=Synthetic())
@@ -109,6 +178,7 @@ class Dataset:
         self.marginals: Optional[Dict[str, Any]] = None
         if marginals is not None:
             self.marginals = json.loads(marginals)
+
         self.policy: Optional[dict] = policy
 
         # We set `unbatch` as the default first transform
@@ -118,24 +188,86 @@ class Dataset:
             transforms.append(("unbatch", {}))
         self.transforms = transforms
 
-    def _add_transform(self, transform: Tuple[str, Dict]) -> Dataset:
-        """Create a new dataset with an additional transform."""
-        return Dataset(
-            slugname=self.name,
-            id=self.id,
-            client=self.client,
-            type_metadata=json.dumps(self.type_metadata),
-            URI=self.URI,
-            human_description=self.human_description,
-            marginals=json.dumps(self.marginals),
-            policy=self.policy,
-            transforms=self.transforms + [transform],
-            synthetic=self._synthetic,
+    @classmethod
+    def _from_dict(cls, data: dict, client: Client) -> Dataset:
+        """Get a dataset from the json data sent by the server.
+
+        Deserializes the protobufs contained in the response and registers them
+        in the storage.
+
+        Args:
+            data (dict): JSON data returned by calling GET /datasets/<id>
+            client (Client): client used to get information from the server
+
+        Returns:
+            Dataset
+        """
+        dataspec_id = data.get("id")
+        serialized_dataspec = data.get("dataspec")
+        slugname = data.get("name")
+
+        if serialized_dataspec is None:
+            # TODO /datasets doesn't return the DataSpec def
+            raise ValueError(f"{slugname} not returned by the server.")
+
+        # Deserialize objects from the JSON response
+        dataspec = (
+            client.context()
+            .factory()
+            .create(dict_deserialize(serialized_dataspec["dataset"]))
+        )
+        client.context().manager().set_remote(dataspec)
+
+        if len(serialized_dataspec["synthetic_dataset"]) == 0:
+            raise ValueError("Synthetic Dataspec not ready")
+        _ = (
+            client.context()
+            .factory()
+            .create(
+                dict_deserialize(serialized_dataspec["synthetic_transform"])
+            )
+        )
+        syn_dataset = (
+            client.context()
+            .factory()
+            .create(dict_deserialize(serialized_dataspec["synthetic_dataset"]))
+        )
+        variant_constraint(
+            constraint_kind=st.ConstraintKind.SYNTHETIC,
+            dataspec=syn_dataset,
+        )
+        attach_variant(
+            dataspec,
+            syn_dataset,
+            kind=st.ConstraintKind.SYNTHETIC,
+        )
+        _ = (
+            client.context()
+            .factory()
+            .create(dict_deserialize(serialized_dataspec["schema"]))
+        )
+        _ = (
+            client.context()
+            .factory()
+            .create(dict_deserialize(serialized_dataspec["synthetic_schema"]))
         )
 
+        return cls(
+            id=dataspec_id,
+            dataspec=dataspec,
+            client=client,
+            type_metadata=data.get("type_metadata"),
+            marginals=data.get("marginals"),
+            human_description=data.get("human_description"),
+            is_bigdata=data.get("is_bigdata"),
+        )
+
+    def __getitem__(self, item):
+        return self.table(item)
+
     def _split(
-        self, split: Union[str, List[str]]
-    ) -> Union[Dataset, List[Dataset]]:
+        self, split: t.Union[str, List[str]]
+    ) -> t.Union[Dataset, List[Dataset]]:
         """Get subsets of the Sarus dataset.
 
         The `split` parameter is a string that specifies a slice of the dataset
@@ -232,21 +364,31 @@ class Dataset:
 
         Returns:
             Dict[str, Dict]: A dictionary holding metadata where each key
-            is a feature.
+            is a table name and each value is a dict with features.
         """
         if self.type_metadata is None:
             return None
-        features: Dict[str, Dict] = self.type_metadata["features"]
+        if len(self.type_metadata) > 1:
+            features: Dict[str, Dict] = {
+                table["name"]: table["features"]
+                for table in self.type_metadata
+            }
+        else:
+            features: Dict[str, Dict] = self.type_metadata[0]["features"]
         return features
 
     def __repr__(self) -> str:
         return f"<Sarus Dataset slugname={self.name} id={self.id}>"
 
+    def _ipython_display_(self) -> None:
+        """Custom ipython display."""
+        print(repr(self))
+
     def as_tensorflow(
         self,
         max_download_size: Optional[int] = None,
         original: bool = False,
-    ) -> sarus.tensorflow.Dataset:
+    ) -> sltf.Dataset:
         """Return the corresponding `sarus.tensorflow.Dataset` object.
 
         This allows to manipulate the Sarus `Dataset` as a Tensorflow dataset.
@@ -266,68 +408,75 @@ class Dataset:
                 will encode categories as integers.
 
         Returns:
-            A sarus.tensorflow.Dataset.
+            A sarus_tensorflow.Dataset.
         """
-        return sarus.tensorflow.Dataset(
+        logging.warning(
+            "'as_tensorflow()' uses the old API, not preprocessing."
+        )
+        if self.is_bigdata:
+            raise Exception(
+                "`as_tensorflow()` is not supported for this dataset"
+            )
+        return sltf.Dataset(
             self,
             max_download_size=max_download_size,
             original=original,
         )
 
-    def as_pandas(
-        self,
-        max_download_size: Optional[int] = None,
-        original: bool = True,
-    ) -> spd.DataFrame:
-        """Return the corresponding `sarus.pandas.DataFrame` object.
+    def as_pandas(self) -> DataFrame:
+        if self.is_bigdata:
+            raise Exception("`as_pandas()` is not supported for this dataset")
+        return DataFrame.from_dataspec(self._dataspec)
 
-        This allows to manipulate the Sarus `Dataset` as a Pandas DataFrame.
+    def table(self, table_name: t.List[str]) -> Table:
+        """Get a table from the dataset.
 
         Args:
-            max_download_size (int, optional): Max number of synthetic data rows
-                to download locally. Indicates the number of synthetic data
-                rows to download from the Sarus server. It will not download
-                more than the maximum number of available synthetic data. If
-                `None`, it will download all the synthetic data. If different
-                from `None`, all local computations will be done on the local
-                synthetic sample so local results will differ from remote
-                results.
-
-            original (bool): Returns categories original values.
-                If True will return categories as original values. If False,
-                will encode categories as integers.
-
+            table_name (List[str]): Name of a form ['namespace_1','namespace_2'...,'table_name'].
+                One can omit namespaces if table names are not ambiguous.
         Returns:
-            A sarus.pandas.DataFrame.
+            Table: Table fitting the given name.
         """
 
-        return spd.DataFrame(
-            self,
-            max_download_size=max_download_size,
-            original=original,
+        dataspec_tables = [
+            el.to_strings_list()[0] for el in self._dataspec.schema().tables()
+        ]
+        # remove protection name and add schema_name
+        dataspec_tables = [
+            [self._dataspec.schema().name(), *(val for val in el[1:])]
+            for el in dataspec_tables
+        ]
+
+        possible_tables_indices = []
+        # iterate over tables and select_ones that have same path
+        n_elements = len(table_name)
+        for i, table_path in enumerate(dataspec_tables):
+            if table_path[-n_elements:] == table_name:
+                possible_tables_indices.append(i)
+
+        try:
+            assert len(possible_tables_indices) == 1
+        except AssertionError:
+            if len(possible_tables_indices) == 0:
+
+                raise ValueError(f"Table {table_name} not found")
+            else:
+                raise ValueError(f"Ambiguous table name {table_name}")
+
+        else:
+            table_path = self._dataspec.schema().tables()[
+                possible_tables_indices[0]
+            ]
+
+        ds_filter = self._dataspec.schema().data_type().get(table_path)
+        filtered_ds = sarus_data_spec.transform.filter(filter=ds_filter)(
+            self._dataspec
+        )
+        selected_ds = sarus_data_spec.transform.get_item(path=table_path)(
+            filtered_ds
         )
 
-    @classmethod
-    def _from_dict(cls, data: dict, client: "Client") -> "Dataset":
-        """Get a dataset from the json data sent by the server.
-
-        Args:
-            data (dict): json data returned by the server
-
-            client (Client): client used to get information from the server
-
-        Returns:
-            Dataset
-        """
-        return cls(
-            data.get("name"),
-            data.get("id"),
-            client,
-            type_metadata=data.get("type_metadata"),
-            marginals=data.get("marginals"),
-            URI=data.get("URI"),
-            human_description=data.get("human_description"),
-        )
+        return Table(selected_ds, table_name)
 
     @property
     def epsilon(self) -> float:
@@ -372,6 +521,7 @@ class Dataset:
         rows_number: Optional[int] = None,
         force_refresh: bool = False,
         original: bool = True,
+        table_name: Optional[str] = None,
     ) -> pa.Table:
         """Fetch synthetic data as a pyarrow.Table.
 
@@ -383,6 +533,9 @@ class Dataset:
             force_refresh (bool): if True, always fetch from server
             original (bool): if False, get categorical values as integers
         """
+        if table_name is None and len(self.type_metadata) > 1:
+            raise ValueError("Table name needs to be specified in this case.")
+
         dataset_size = int(self.marginals["rows"])
         if rows_number is None:
             rows_number = dataset_size
@@ -412,8 +565,12 @@ class Dataset:
 
         if not has_enough_synthetic or force_refresh:
             # Fetch synthetic data
+            uri = f"{self.client.base_url}/synthetic_data/{self.id}"
+            if table_name is not None:
+                uri = f"{uri}/{table_name}"
+
             resp = self.client.session.get(
-                f"{self.client.base_url}/synthetic_data/{self.id}",
+                uri,
                 stream=True,
                 params={
                     "textual_categories": original,
@@ -437,6 +594,7 @@ class Dataset:
         rows_number: Optional[int] = None,
         force_refresh: bool = False,
         original: bool = True,
+        table_name: Optional[str] = None,
     ):
         """Return synthetic data as a pandas.DataFrame.
 
@@ -448,6 +606,9 @@ class Dataset:
         Returns:
             pandas.DataFrame: synthetic data
         """
+        if table_name is None and len(self.type_metadata) > 1:
+            raise ValueError("Table name needs to be specified in this case.")
+
         synthetic_data = self._fetch_synthetic(
             rows_number=rows_number,
             force_refresh=force_refresh,
@@ -457,7 +618,12 @@ class Dataset:
         synthetic_df = synthetic_data.to_pandas()
 
         # Convert images encoded as bytes to 3d arrays
-        for feature in self.features:
+        if table_name is None:
+            features = self.features
+        else:
+            features = self.features[table_name]
+
+        for feature in features:
             dtype = list(feature["type"].keys())[0]
             if dtype == "image":
                 synthetic_df[feature["name"]] = synthetic_df[
@@ -591,6 +757,7 @@ class Dataset:
         rows_number: Optional[int] = None,
         original: bool = False,
         force_refresh: bool = False,
+        table_name: Optional[str] = None,
     ) -> tf.data.Dataset:
         """Return synthetic data as a tensorflow.data.Dataset.
 
@@ -603,6 +770,9 @@ class Dataset:
         Returns:
             tensorflow.data.Dataset: synthetic data
         """
+        if table_name is None and len(self.type_metadata) > 1:
+            raise ValueError("Table name needs to be specified in this case.")
+
         if force_refresh:
             self._fetch_synthetic(
                 force_refresh=force_refresh,
@@ -610,6 +780,11 @@ class Dataset:
                 rows_number=rows_number,
             )
 
+        if table_name is None:
+            features = self.features
+        else:
+            features = self.features[table_name]
+
         # Generator function iterating pyarrow RecordBatches
         def generator() -> Dict[str, List[Any]]:
             synthetic_table = self._fetch_synthetic(
@@ -618,13 +793,11 @@ class Dataset:
             for batch in synthetic_table.to_batches(max_chunksize=batch_size):
                 yield Dataset._adapt_for_tf(
                     batch=batch.to_pydict(),
-                    features=self.features,
+                    features=features,
                     original=original,
                 )
 
-        tf_signature = Dataset._sarus_features_to_tf_spec(
-            self.features, original
-        )
+        tf_signature = Dataset._sarus_features_to_tf_spec(features, original)
 
         return tf.data.Dataset.from_generator(
             generator, output_signature=tf_signature
@@ -822,6 +995,9 @@ class Dataset:
             escape=False,
         )
 
+    def tables(self):
+        return _get_tables(self._dataspec)
+
 
 class Client:
     """Entry point for the Sarus API client."""
@@ -864,14 +1040,78 @@ class Client:
         else:
             self._credentials_login(email, password)
 
+        self._context = LocalSDKContext(self)
+        self._context.set_sync_policy(SyncPolicy.SEND_ON_VALUE)
+        push_global_context(self._context)
+
+    def context(self) -> LocalSDKContext:
+        return self._context
+
+    def save(self, wrapper: DataSpecWrapper, path: str) -> None:
+        """Save the DataSpecWrapper for loading in another notebook."""
+        manager: LocalSDKManager = self.context().manager()
+
+        user_ds = wrapper.dataspec(DataSpecVariant.USER_DEFINED)
+        alternative_ds = wrapper.dataspec(DataSpecVariant.ALTERNATIVE)
+        mock_ds = wrapper.dataspec(DataSpecVariant.MOCK)
+
+        user_graph = manager.computation_graph(user_ds)
+        mock_graph = manager.computation_graph(mock_ds)
+
+        user_refs = (
+            set(user_graph["dataspecs"])
+            | set(user_graph["transforms"])
+            | set(user_graph["attributes"])
+            | set(user_graph["variant_constraints"])
+        )
+
+        alternative_refs = {alternative_ds}
+
+        mock_refs = (
+            set(mock_graph["dataspecs"])
+            | set(mock_graph["transforms"])
+            | set(mock_graph["attributes"])
+            | set(mock_graph["variant_constraints"])
+        )
+
+        refs = user_refs | alternative_refs | mock_refs
+        serialized_refs = [serialize(ref.protobuf()) for ref in refs]
+
+        with open(path, "wb") as f:
+            pkl.dump(
+                {
+                    "serialized_obj": serialized_refs,
+                    "user_defined": user_ds.uuid(),
+                },
+                f,
+            )
+
+    def load(self, path: str) -> DataSpecWrapper:
+        """Load a DataSpecWrapper from a file."""
+        with open(path, "rb") as f:
+            data = pkl.load(f)
+
+        refs = data["serialized_obj"]
+        user_defined = data["user_defined"]
+
+        factory = self.context().factory()
+        storage = self.context().storage()
+        # Two passes to have all the correct referrings
+        refs = [factory.create(deserialize(ref)) for ref in refs]
+        _ = [storage.store(ref) for ref in refs]
+
+        user_ds = self.context().storage().referrable(user_defined)
+
+        return self.context().wrapper_factory().create(user_ds)
+
     def _oidc_login(self):
         oidc_login_url = f"{self.base_url}/oidc_login?headless=true"
         try:
             from IPython.display import Javascript, clear_output
 
-            display(  # noqa: F821
+            display(
                 Javascript(f'window.open("{oidc_login_url}");')
-            )
+            )  # noqa: F821
             display(clear_output())  # noqa: F821
         except Exception:
             webbrowser.open(oidc_login_url)
@@ -924,22 +1164,23 @@ class Client:
         request = self.session.get(f"{self.base_url}/datasets")
         return [ds["name"] for ds in request.json()]
 
-    def list_datasets(self) -> List[Dataset]:
+    def list_datasets(self) -> DatasetList:
         """List available Sarus datasets.
 
         Returns:
-            List[Dataset]: List of available Sarus datasets.
+            DatasetList: List of available Sarus datasets.
         """
         request = self.session.get(f"{self.base_url}/datasets")
         unsorted_datasets = [
-            Dataset._from_dict(ds_json, self)
+            ShallowDataset._from_dict(ds_json)
             for ds_json in request.json()
             if ds_json["status"] == "ready"
         ]
-        return sorted(
+        sorted_datasets = sorted(
             unsorted_datasets,
             key=lambda x: x.name,
         )
+        return DatasetList(sorted_datasets, self)
 
     def dataset(self, slugname: str = None, id: int = None) -> Dataset:
         """Select a dataset from the Sarus Gateway.
@@ -997,7 +1238,7 @@ class Client:
     def query(
         self,
         query: str,
-        target_epsilon: float = 0.0,
+        target_epsilon: Optional[float] = None,
         verbose: bool = True,
         use_old_query=False,
     ) -> int:
@@ -1006,9 +1247,15 @@ class Client:
         Args:
             query (String): SQL query
 
-            target_epsilon (float): Maximum privacy budget (epsilon) to assign to the query.
-                If 0, runs on the synthetic data. If >0, results are a combination of a
-                query on the synthetic data and a differentially-private query on the real data.
+            target_epsilon (Optional[float]): Maximum privacy budget (epsilon) to assign to the query.
+                If 0, runs on the synthetic data.
+                If >0, result is a combination of a query on the synthetic data and a
+                Differentially-Private query on the real data.
+                If None, a default target epsilon specific to the current user and access rule is used.
+                Default target epsilon is 0 if the access is a Differentially-Private access with
+                per-user or per-group limit; default value equals to per-query limit if the access is
+                a Differentially-Private access with a per-query limit only. Meaning Sarus maximizes result
+                accuracy in the given privacy constraints. See user documentation to know more.
 
             use_old_query (bool): Whether to use the v1 version of the query task.
 
@@ -1018,12 +1265,16 @@ class Client:
         endpoint = "query"
         if use_old_query:
             endpoint = "query_v1"
+
+        payload = {
+            "query": query,
+        }
+        if target_epsilon is not None:
+            payload["target_epsilon"] = target_epsilon
+
         request = self.session.post(
             f"{self.base_url}/{endpoint}",
-            json={
-                "query": query,
-                "target_epsilon": target_epsilon,
-            },
+            json=payload,
         )
         if request.status_code > 200:
             if request.status_code == 403:
@@ -1057,7 +1308,38 @@ class Client:
                 f"Actual privacy consumption (epsilon): "
                 f"{ds.epsilon-start_eps:.03f}"
             )
-        return status
+
+        if status["status"] == "SUCCESS":
+            status_with_pyobjs = self._convert_to_result_with_pyobjs(status)
+
+            return status_with_pyobjs
+        else:
+            return status
+
+    @staticmethod
+    def _convert_to_result_with_pyobjs(
+        status: Dict[str, Any]
+    ) -> Dict[str, Any]:
+        """Convert the query result values to native Python objects using the
+           description field.
+
+        Args:
+            status (dict): the status response for a given completed query task returned by the server.
+
+        Returns:
+            dict: a new status dict with the initialized Python objects.
+        """
+        result = []
+        for row in status["result"]:
+            for idx, col_desc in enumerate(status["description"]):
+                _, class_name, _, _, _, _, _ = col_desc
+                if class_name == str(decimal.Decimal):
+                    row[idx] = decimal.Decimal(row[idx])
+            result.append(row)
+        return {
+            **status,
+            "result": result,
+        }
 
     def _abort_training(self, id: int):
         """Abort a training on the Sarus Gateway.
@@ -1183,6 +1465,20 @@ class Client:
             return tf.keras.models.load_model(_dir)
 
 
+class Table:
+    """Class representing a Sarus Table, part of a Sarus Dataset."""
+
+    def __init__(self, filtered_dataspec, table_name: t.List[str]):
+        self._dataspec = filtered_dataspec
+        self.name = table_name
+
+    def __repr__(self) -> str:
+        return f"<Sarus Table name={self.name}>"
+
+    def as_pandas(self) -> DataFrame:
+        return DataFrame.from_dataspec(self._dataspec)
+
+
 def _save_model(model: tf.keras.Model) -> bytes:
     """Convert a keras Model to compressed archive format."""
     with tempfile.TemporaryDirectory() as _dir:
@@ -1194,3 +1490,17 @@ def _save_model(model: tf.keras.Model) -> bytes:
             with open(path, "rb") as f:
                 ret = f.read()
                 return ret
+
+
+def _get_tables(dataspec) -> t.List[t.List[str]]:
+    """For given parameters of a Sarus Dataset return a list of Sarus Tables."""
+
+    dataspec_tables = [
+        el.to_strings_list()[0] for el in dataspec.schema().tables()
+    ]
+    # remove protection name and add schema_name
+    dataspec_tables = [
+        [dataspec.schema().name(), *(val for val in el[1:])]
+        for el in dataspec_tables
+    ]
+    return dataspec_tables
